""" A set of constants to describe the package.

Don't put any code in here, because it must be safe to execute in setup.py. """

__title__ = 'space_tracer'  # => name in setup.py
__version__ = '4.0.0'
__author__ = "Don Kirkby"
__author_email__ = "donkirkby@gmail.com"
__description__ = "Trade time for space when debugging your code."
__url__ = "https://donkirkby.github.io/live-py-plugin/"
